#include "Scratchpad.h"

#include "../pointers/pointers.h"
#include <stdlib.h>
#include <string.h>

int Scratchpad_Create(Scratchpad* target, size_t capacity, allocator_t* allocator)
{
	if (target == NULL) return EDESTADDRREQ;

	void* memory = Allocator_Allocate(allocator, capacity);
	if (memory == NULL) return ENOMEM;

	target->memory = memory;
	target->capacity = capacity;
	target->reserved = 0;
	target->next     = NULL;
	target->allocator = allocator;

	return EXIT_SUCCESS;
}

static void* _Scratchpad_ReserveHere(Scratchpad* pad, size_t size)
{
	if (size > pad->capacity - pad->reserved) return NULL;

	void* scratch_buffer = advancep(pad->memory, pad->reserved);
	pad->reserved += size;

	return scratch_buffer;
}

static int _Scratchpad_CreateEmbedded(Scratchpad* pad, size_t size)
{
	Scratchpad* new = Allocator_Allocate(pad->allocator, sizeof(*pad));
	if (new != NULL) {
		// Copy current pad in the *next pointer
		memcpy(new, pad, sizeof(*pad));

		// Calculate new capacity
		size_t new_capacity = size > pad->capacity ?
			pad->capacity * (size / pad->capacity + 1)
			: pad->capacity;

		// Initialize new pad
		if (Scratchpad_Create(pad, new_capacity, pad->allocator)) {
			Allocator_Free(pad->allocator, new, sizeof(*new));
			return EXIT_FAILURE;
		} else {
			pad->next = new;
		}
	} else {
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

void* Scratchpad_Reserve(Scratchpad* pad, size_t size)
{
	if (pad == NULL) return NULL;

	Scratchpad* current = pad;
	void* reserve = NULL;
	while (reserve == NULL && current != NULL) {
		reserve = _Scratchpad_ReserveHere(current, size);
		current = current->next;
	}

	if (reserve == NULL) {
		if (_Scratchpad_CreateEmbedded(pad, size)) return NULL;
		reserve = _Scratchpad_ReserveHere(pad, size);
	}

	return reserve;
}

void* Scratchpad_ReserveAligned(Scratchpad* pad, size_t amount, size_t alignment)
{
	if (pad == NULL) return NULL;

	if (alignment == 1) return Scratchpad_Reserve(pad, amount);

	Scratchpad* root = pad;
	do {
		uintptr_t pointer_value     = (uintptr_t) pad->memory;
		uintptr_t alignment_padding = alignment - (pointer_value % alignment);
		void* memory = _Scratchpad_ReserveHere(pad, alignment_padding + amount);
		if (memory != NULL) return memory;

		pad = pad->next;

	} while (pad != NULL);

	
	if (_Scratchpad_CreateEmbedded(root, amount)) return NULL;

	return _Scratchpad_ReserveHere(root, amount);
}

int Scratchpad_Reclaim(Scratchpad* pad, void* pointer, size_t length)
{
	if (pad == NULL) return EDESTADDRREQ;
	if (pointer == NULL) return EINVAL;
	if (length > pad->reserved) return EBOUNDS;

	if (rewindp(advancep(pad->memory, pad->reserved), length) == pointer) {
		pad->reserved -= length;
	}

	return EXIT_SUCCESS;
}

static void _Scratchpad_DestroyHere(Scratchpad* pad)
{
	Allocator_Free(pad->allocator, pad->memory, pad->capacity);

	return;
}

void Scratchpad_Reset(Scratchpad* pad)
{
	if (pad == NULL) return;

	while (pad->next != NULL && pad->next->next != NULL) {
		Scratchpad* current = pad->next;
		pad->next = current->next;

		_Scratchpad_DestroyHere(current);
		Allocator_Free(pad->allocator, current, sizeof(*current));
	}

	if (pad->next != NULL) {
		Scratchpad* next = pad->next;
		_Scratchpad_DestroyHere(pad);
		memcpy(pad, next, sizeof(*next));

		Allocator_Free(pad->allocator, next, sizeof(*next));
	}

	return;
}

void Scratchpad_Destroy(Scratchpad* pad)
{
	if (pad == NULL) return;

	while (pad->next != NULL) {
		Scratchpad* next_next = pad->next->next;
		_Scratchpad_DestroyHere(pad->next);
		Allocator_Free(pad->allocator, pad->next, sizeof(*pad));

		pad->next = next_next;
	}

	_Scratchpad_DestroyHere(pad);

	return;
}
