/*
 * This code is part of the strategy game operational-space.
 * operational-space comes with ABSOLUTELY NO WARRANTY and is licensed under GPL-2.0.
 *    Copyright (C) 2024  VegOwOtenks, Sleppo04
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "allocator-interface.h"


// Forwarding
void* Allocator_ForwardArrayAllocation(allocator_t* allocator, size_t object_size, size_t length)
{
	return allocator->allocate(allocator, object_size * length);
}

int Allocator_ForwardArrayFree(allocator_t* allocator, void* pointer, size_t object_size, size_t length)
{
	return allocator->free(allocator, pointer, object_size * length);
}

void* Allocator_ForwardReallocate(allocator_t* allocator, void* pointer, size_t old_size, size_t new_size)
{
	void* new = allocator->allocate(allocator, new_size);
	if (new == NULL) {
		return NULL;
	}

	memcpy(new, pointer, old_size);

	if (allocator->free(allocator, pointer, old_size)) {
		allocator->free(allocator, new, new_size);
		return NULL;
	}

	return new;
}

// Wrapping System
int Allocator_WrapSystemFree(allocator_t* allocator, void* pointer, size_t size)
{
	free(pointer);

	allocator->reserved -= size;

	return EXIT_SUCCESS;
}

void* Allocator_WrapSystemAllocate(allocator_t* allocator, size_t size)
{
	void* result = malloc(size);

	if (result != NULL) {
		allocator->reserved += size;
	}

	return result;
}


void* Allocator_WrapSystemReallocate(allocator_t* allocator, void* pointer, size_t old_size, size_t new_size)
{
	void* result = realloc(pointer, new_size);

	if (result != NULL) {
		allocator->reserved -= old_size;
		allocator->reserved += new_size;
	}

	return result;
}


void* Allocator_Allocate(allocator_t* allocator, size_t size)
{
	if (allocator == NULL) {
		return malloc(size);
	}

	return allocator->allocate(allocator, size);
}

int   Allocator_Free(allocator_t* allocator, void* pointer, size_t size)
{
    if (pointer == NULL) {
        return EXIT_SUCCESS;
    }
	if (allocator == NULL) {
		free(pointer);
		return EXIT_SUCCESS;
	}

	return allocator->free(allocator, pointer, size);
}

void* Allocator_AllocateArray(allocator_t* allocator, size_t object_size, size_t length)
{
	if (allocator == NULL) {
		return calloc(length, object_size);
	}

	return allocator->allocateArray(allocator, object_size, length);
}

int Allocator_FreeArray(allocator_t* allocator, void* pointer, size_t object_size, size_t length)
{
	if (allocator == NULL) {
		free(pointer);
		return EXIT_SUCCESS;
	}

	return allocator->freeArray(allocator, pointer, object_size, length);
}

void* Allocator_Reallocate(allocator_t* allocator, void* pointer, size_t old_size, size_t new_size)
{
	if (allocator == NULL) {
		return realloc(pointer, new_size);
	}

	return allocator->reallocate(allocator, pointer, old_size, new_size);
}

int Allocator_CreateSystemAllocator(allocator_t* destination)
{
    destination->xdata = NULL;
    destination->reserved      = 0;

    destination->allocateArray = Allocator_ForwardArrayAllocation;
    destination->freeArray     = Allocator_ForwardArrayFree;
    destination->free          = Allocator_WrapSystemFree;
    destination->allocate      = Allocator_WrapSystemAllocate;
    destination->reallocate    = Allocator_WrapSystemReallocate;

    return EXIT_SUCCESS;
}

void Allocator_DestroySystemAllocator(allocator_t* allocator)
{
	memset(allocator, 0, sizeof(allocator[0]));
	return;
}
